\Grinder.xpl	30-Jun-2020
\Rotatable 3D Model of a Moving Elliptic Trammel - Better Known as a BS Grinder
\
\Revisions:
\10-Mar-2014: Modified for RPi.
\04-Oct-2015: Modified for new GetMouse intrinsic, and EraseImage with FillMem.
\30-Jun-2020: Reverse GetMouseMove Y axis for consistency.

def	VidMode=$105,		\video mode: 8-bit color graphics
	ScrW=1024, ScrH=768,	\screen width and height in pixels
	ImW=ScrH, ImH=ScrH;	\image width and height in pixels (a square)

int	XRay;			\minimum Z distance that can be seen (xray view)
char	Image(ImW*ImH);		\image buffer, gets copied to the graphic screen
int	ZBuffer(ImW*ImH);	\depth (Z distance) of each pixel in Image

def	Bel=$07, CR=$0D, Esc=$1B;	\control characters



proc	Fatal(Msg);		\Display error message and terminate program
char	Msg;
[SetVid(3);			\restore text mode
Text(0, Msg);
exit;
];	\Fatal



proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
[V3(0):= V1(1)*V2(2) - V1(2)*V2(1);
 V3(1):= V1(2)*V2(0) - V1(0)*V2(2);
 V3(2):= V1(0)*V2(1) - V1(1)*V2(0);
];	\CrossProd



proc	Rotate(V, R, P, W);	\Rotate vector in 3 dimensions
real	V,		\3D vector, a 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	SW, SP, SR, CW, CP, CR, T;
[SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
 CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= V(1)*CR - V(2)*SR;		\		       	 
V(2):= V(2)*CR + V(1)*SR;	\		      1   0   0 
V(1):= T;			\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= V(0)*CP + V(2)*SP;		\		       	 
V(2):= V(2)*CP - V(0)*SP;	\		      CP  0  -SP
V(0):= T;			\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= V(0)*CW - V(1)*SW;		\		       	  
V(1):= V(1)*CW + V(0)*SW;	\		       CW  SW	0 
V(0):= T;			\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
];	\Rotate					       	  



proc	EraseImage;		\Erase Image array and initialize ZBuffer
[FillMem(Image, $0F, ImW*ImH);	\black
FillMem(ZBuffer, $7F, ImW*ImH*4); \init to nearly maximum depth: $7F7F_7F7F
];	\EraseImage

\-------------------------------------------------------------------------------

proc	DrawTriangle(Tri);	\Draw a colored triangle onto the Image array
int	Tri;			\array of three points(x,y,z) and one color
int	Top, Bot,		\top and bottom Y coordinates
	YImW;			\Y coordinate times Image width in pixels
int	Color, I, Z, Slope5;	\Slope5 = slope in X-Z plane
int	Z3, Z4, Y, T;		\Z depths along left and right triangle edges
int	Edge1, Edge2;		\X coords along left & right edges of triangles
real	REdge1, REdge2,		\real versions accumulate fractional pixels
	Slope1, Slope2;		\slopes of left and right edges in X-Y plane
int	Slope3, Slope4;		\slopes of left and right edges in X-Z plane
int	X0,Y0,Z0, X1,Y1,Z1, X2,Y2,Z2; \non-indexed Tri coordinates, for speed


\                             X0,Y0
\                              /.  . Top
\                            //               Right-handed coordinate system
\                          / /
\              Slope1    /  /
\                      /   /                            |   Z (into screen)
\            Edge1   /    /                             |  /
\                  /     /  Slope2                      | /
\                /      /                               |/
\              /       /                        --------+-------- X
\            /        /  Edge2                         /|
\    X1,Y1 /_ _ _ _ _/ .  .  .  .  . Bot              / |
           \        /                Top             /  |
            \      /                                    |
             \    /                                     Y
              \  /
               \/.  .  .  .  .  .  . Bot
\            X2,Y2


	proc	DrawHalf;	\(optimized procedure with no local variables)
	[if Bot > ImH then Bot:= ImH-1;			\clip to bottom of Image
	YImW:= Top*ImW;
	for Y:= Top to Bot-1 do				\from top to bottom...
		[if Y >= 0 then				\clip to top of Image
			[Edge1:= fix(REdge1);		\get integer X coords
			Z:= Z3;				\depth at left edge
			Edge2:= fix(REdge2);
			if Edge1 > Edge2 then		\order left to right
				[T:= Edge1;  Edge1:= Edge2;  Edge2:= T;
				Z:= Z4;			\depth at right edge
				];
			if Edge1 < 0 then
				[Z:= Z - Edge1*Slope5;
				Edge1:= 0;		\clip to left side
				];
			if Edge2 > ImW then Edge2:= ImW;\clip to right side
			for I:= Edge1+YImW to Edge2-1+YImW do
			    [if Z<ZBuffer(I) & Z>XRay then
				[Image(I):= Color;	\fill scanline
				 ZBuffer(I):= Z;	\draw pixel if closer
				];
			    Z:= Z + Slope5;		\adjust depth
			    ];
			];
		REdge1:= REdge1 + Slope1;		\adj X coords according
		Z3:= Z3 + Slope3;			\ to slopes
		REdge2:= REdge2 + Slope2;
		Z4:= Z4 + Slope4;
		YImW:= YImW + ImW;			\index for next scanline
		];
	];	\DrawHalf

[	\DrawTriangle
X0:= Tri(0,0);			\get vertices
X1:= Tri(1,0);
X2:= Tri(2,0);

if X0>=ImW & X1>=ImW & X2>=ImW then return;		\clip entire triangle
if X0<0 & X1<0 & X2<0 then return;

Y0:= Tri(0,1);
Y1:= Tri(1,1);
Y2:= Tri(2,1);

Z0:= Tri(0,2);
Z1:= Tri(1,2);
Z2:= Tri(2,2);

if Y0 > Y1 then			\sort vertices from top to bottom (Y0<=Y1<=Y2)
	[T:= Y1;  Y1:= Y0;  Y0:= T;
	 T:= X1;  X1:= X0;  X0:= T;
	 T:= Z1;  Z1:= Z0;  Z0:= T;
	];
if Y0 > Y2 then
	[T:= Y2;  Y2:= Y0;  Y0:= T;
	 T:= X2;  X2:= X0;  X0:= T;
	 T:= Z2;  Z2:= Z0;  Z0:= T;
	];
if Y0 >= ImH then return;				\clip entire triangle
if Y1 > Y2 then
	[T:= Y2;  Y2:= Y1;  Y1:= T;
	 T:= X2;  X2:= X1;  X1:= T;
	 T:= Z2;  Z2:= Z1;  Z1:= T;
	];
if Y2 < 0 then return;					\clip entire triangle

if Y1 # Y0 then
	[Slope1:= float(X1-X0) / float(Y1-Y0);
	 Slope3:= (Z1-Z0) / (Y1-Y0);
	];

if Y2 # Y0 then
	[Slope2:= float(X2-X0) / float(Y2-Y0);
	 Slope4:= (Z2-Z0) / (Y2-Y0);
	 if Y1#Y0 \Slope1 is defined\ & Slope2#Slope1 \works here\ then
	 	Slope5:= fix(float(Slope4-Slope3) / (Slope2-Slope1));
	];
REdge1:= float(X0);  REdge2:= REdge1;
Top:= Y0;  Bot:= Y1;
Color:= Tri(3);
Z3:= Z0;  Z4:= Z0;
DrawHalf;

if Y2 # Y1 then
	[Slope1:= float(X2-X1) / float(Y2-Y1);
	 Slope3:= (Z2-Z1) / (Y2-Y1);
	 if Slope2 # Slope1 then	\in case Slope5 is not defined above
	 	Slope5:= fix(float(Slope4-Slope3) / (Slope2-Slope1));
	];
REdge1:= float(X1);			\(in case X1 = X0)
Top:= Y1;  Bot:= Y2;
Z3:= Z1;
DrawHalf;
];	\DrawTriangle

\######################### Display and Move the Scene ##########################

proc	MoveScene;		\Move and draw the scene
def	Blocks = 17+1;		\*** total number of blocks (see MakeScene)
def	FacesMax = Blocks*6;	\maximum number of faces
def	VertsMax = Blocks*8;	\maximum number of vertices
char	FaceColor(FacesMax);	\color of each face
int	FaceVert0(FacesMax), FaceVert1(FacesMax),	\coords of each corner
	FaceVert2(FacesMax), FaceVert3(FacesMax);	\ of each face
real	VertexX(VertsMax), VertexY(VertsMax), VertexZ(VertsMax);
int	RotVertX(VertsMax),	\rotated vertices, ready to be drawn
	RotVertY(VertsMax),
	RotVertZ(VertsMax);

\-------------------------------------------------------------------------------

proc	DrawScene;		\Draw the scene onto Image array
\Inputs: Blocks; FaceColor; FaceVert0,1,2,3; RotVertX,Y,Z;
int	F, B, Z0, Z2, Cz, J,
	I0, I1, I2, I3;
int	Tri(4,3);
[F:= 0;	\face
\Cz = distance (*2) of the center the block
for B:= 0 to Blocks-1 do
	[Z0:= RotVertZ(B*8+0);		\V0(2), Z distance of corner
	 Cz:= Z0 + RotVertZ(B*8+6);	\plus Z of opposite corner

\	 1     0	each face is drawn as two triangles
\	  +---+
\	  |  /:
\	  | / :
\	  + - +
\	 2     3

	for J:= 0 to 6-1 do		\for all the faces of the block...
		[I0:= FaceVert0(F);	\get distances of opposite corners
		Z0:= RotVertZ(I0);
		I2:= FaceVert2(F);
		Z2:= RotVertZ(I2);

		\if face is farther than center of its block then don't draw it
		if Z2+Z0 <= Cz+1 then
		       [Tri(0,0):= RotVertX(I0);	\for all 3 dimensions
			Tri(0,1):= RotVertY(I0);
			Tri(0,2):= Z0;

			I1:= FaceVert1(F);
			Tri(1,0):= RotVertX(I1);
			Tri(1,1):= RotVertY(I1);
			Tri(1,2):= RotVertZ(I1);

			Tri(2,0):= RotVertX(I2);
			Tri(2,1):= RotVertY(I2);
			Tri(2,2):= Z2;

			Tri(3):= FaceColor(F);
			DrawTriangle(Tri);

			I3:= FaceVert3(F);
			Tri(1,0):= RotVertX(I3);
			Tri(1,1):= RotVertY(I3);
			Tri(1,2):= RotVertZ(I3);
			DrawTriangle(Tri);
		       ];
		F:= F+1;				\next face
		];
	];
];	\DrawScene

\-------------------------------------------------------------------------------

real	Zoom,			\magnification factor
	Ref0(3), Ref1(3), Ref2(3); \rotated reference frame (unit vectors)
int	Offset(2);		\scene's panned position vector

proc	RotateScene;		\Rotate, pan and zoom scene using the mouse
int	I;			\Inputs: VertexX,Y,Z. Outputs: RotVertX,Y,Z
int	Mouse, Del;		\mouse arguments
real	AngX, AngY, AngZ,	\rotation angle about indicated axis (radians)
	Ref00, Ref01, Ref02,	\Ref0(0) etc. sans index, for speed
	Ref10, Ref11, Ref12,
	Ref20, Ref21, Ref22,
	Rot0, Rot1, Rot2,
	Dist, Pers;		\Z distance modified by perspective
int	OffX, OffY;
def	CX=ImW/2, CY=ImH/2;	\center coordinates of Image area
[AngY:= 0.0;  AngX:= 0.0;  AngZ:= 0.0;

Mouse:= GetMouse;
Del:= GetMouseMove;
case Mouse(2) & $03 of
  $01:	Zoom:= Zoom + Float(Del(1))/100.0;	\left button
  $02:	[Offset(0):= Offset(0) + Del(0);	\right button
	 Offset(1):= Offset(1) + Del(1);	\pan motion
	];
  $03:	XRay:= XRay - Del(1)*10000		\both buttons
other	[\Mouse X motion rotates scene about Y axis
	AngY:= Float(-Del(0)) / 1000.0;
	\Mouse Y motion rotates scene about X axis
	AngX:= Float(+Del(1)) / 1000.0;
	];
\Rotate reference frame according to changes in mouse position.
Rotate(Ref0, AngX, AngY, AngZ);
Rotate(Ref1, AngX, AngY, AngZ);
CrossProd(Ref0, Ref1, Ref2);		\Ref2:= Ref0 x Ref1

\Rotate each vertex in the scene so that it aligns with the reference frame.
\Multiply each point at its initial position (Po) by the rotation matrix to get
\ its rotated position:
\					     				 
\		     	 		     Ref0(0)  Ref0(1)  Ref0(2) 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ref1(0)  Ref1(1)  Ref1(2) 
\		     	 		     Ref2(0)  Ref2(1)  Ref2(2) 
\					     				 
\
Ref00:= Ref0(0);	Ref01:= Ref0(1);	Ref02:= Ref0(2);
Ref10:= Ref1(0);	Ref11:= Ref1(1);	Ref12:= Ref1(2);
Ref20:= Ref2(0);	Ref21:= Ref2(1);	Ref22:= Ref2(2);

Pers:= 1000.0*Zoom;
OffX:= Offset(0) + CX;
OffY:= Offset(1) + CY;

for I:= 0 to VertsMax-1 do		\for all the vertices (corners)...
       [Rot0:= VertexX(I)*Ref00 + VertexY(I)*Ref10 + VertexZ(I)*Ref20;
	Rot1:= VertexX(I)*Ref01 + VertexY(I)*Ref11 + VertexZ(I)*Ref21;
	Rot2:= VertexX(I)*Ref02 + VertexY(I)*Ref12 + VertexZ(I)*Ref22;
	Dist:= Pers / (Rot2 + 1000.0);			\perspective
        RotVertX(I):= fix(Rot0*Dist) + OffX;		\pan and
	RotVertY(I):= fix(Rot1*Dist) + OffY;		\ zoom
	RotVertZ(I):= fix(Rot2*100_000.0);		\scale up for resolution
	];
];	\RotateScene

\===============================================================================

proc	MoveSliders;		\Move sliders and crank arm VertexX,Y,Z
real	T, Dx, Dy;	\time, displacement
int	V;		\vertex
real	Hy, R1(2), R2(2);
real	RVertexX(8), RVertexY(8);
def	Pi2 = 3.141592654 * 2.0;
[
T:= float(GetTime)/1_000_000.0;	\current time of day in seconds
T:= Mod(T, 10.0);		\current time from 0.0 to 9.999... seconds
Dx:= 50.0 * Sin(T/10.0*Pi2);	\+/- 50.0. Repeats every 10 seconds
Dy:= 50.0 * Cos(T/10.0*Pi2);

for V:= 0 to 8-1 do		\for all the vertices (corners) of a block...
       [VertexY( 5*8+V):= VertexY( 5*8+V) + Dy;		\sliders 5, 6, 15, 16
	VertexX( 6*8+V):= VertexX( 6*8+V) + Dx;
	VertexY(15*8+V):= VertexY(15*8+V) + Dy;
	VertexX(16*8+V):= VertexX(16*8+V) + Dx;
	VertexY( 8*8+V):= VertexY( 8*8+V) + Dy;		\pivots 8, 9
	VertexX( 9*8+V):= VertexX( 9*8+V) + Dx;
	VertexY(10*8+V):= VertexY(10*8+V) - Dy;		\knob 10
	VertexX(10*8+V):= VertexX(10*8+V) + 2.0*Dx;
       ];

\Rotate arm
Hy:= sqrt(Dx*Dx + Dy*Dy);		\hypotenuse
R1(0):= Dx/Hy;	R1(1):= -Dy/Hy;		\2D rotation matrix
R2(0):= Dy/Hy;	R2(1):=  Dx/Hy;
for V:= 0 to 8-1 do			\rotate all the vertices...
       [RVertexX(V):= VertexX(7*8+V)*R1(0) + VertexY(7*8+V)*R2(0);
	RVertexY(V):= VertexX(7*8+V)*R1(1) + VertexY(7*8+V)*R2(1);
       ];
for V:= 0 to 8-1 do			\put temporary RVertex back into array
       [VertexX(7*8+V):= RVertexX(V);
	VertexY(7*8+V):= RVertexY(V) + Dy;
       ];
];	\MoveSliders



int	Faces, Verts;			\accumulated number of faces & vertices

proc	AddBlock(V0, V6, C);		\Add a block to Face and Vertex arrays
\Outputs: FaceColor; FaceVert0,1,2,3; VertexX,Y,Z;
real	V0, V6;		\3D coordinates of diagonally opposite vertices
int	C;		\array of 3 face colors (opposite faces are same color)
real	V1(3), V2(3), V3(3), V4(3), V5(3), V7(3);
def	X, Y, Z;
\
\	    4____5	Block vertex (corner) numbers
\	    /:  /|
\	  0/___/1| 
\	   | 7.|./6
\	   |___|/
\	  3    2

	proc	AddVert(V);		\Add coordinates into Vertex array
	real	V;
	[VertexX(Verts):= V(0);
	 VertexY(Verts):= V(1);
	 VertexZ(Verts):= V(2);
	 Verts:= Verts+1;
	];	\AddVert


	proc	AddFace(I0, I1, I2, I3, C);	\Add indexes into Face array
	int	I0, I1, I2, I3, C;
	[FaceVert0(Faces):= Verts+I0;
	 FaceVert1(Faces):= Verts+I1;
	 FaceVert2(Faces):= Verts+I2;
	 FaceVert3(Faces):= Verts+I3;
	 FaceColor(Faces):= C;
	 Faces:= Faces+1;
	];	\AddFace


[\Store index to coordinates for each of the vertices (corners) of the block
AddFace(0, 1, 2, 3, C(2));
AddFace(1, 5, 6, 2, C(0));
AddFace(5, 4, 7, 6, C(2));
AddFace(4, 0, 3, 7, C(0));
AddFace(3, 2, 6, 7, C(1));
AddFace(4, 5, 1, 0, C(1));

\Store the coordinates for the vertices (corners) of the block
V1(X):= V6(X);	V1(Y):= V0(Y);	V1(Z):= V0(Z);
V2(X):= V6(X);	V2(Y):= V6(Y);	V2(Z):= V0(Z);
V3(X):= V0(X);	V3(Y):= V6(Y);	V3(Z):= V0(Z);
V4(X):= V0(X);	V4(Y):= V0(Y);	V4(Z):= V6(Z);
V5(X):= V6(X);	V5(Y):= V0(Y);	V5(Z):= V6(Z);
V7(X):= V0(X);	V7(Y):= V6(Y);	V7(Z):= V6(Z);

AddVert(V0);  AddVert(V1);  AddVert(V2);  AddVert(V3);
AddVert(V4);  AddVert(V5);  AddVert(V6);  AddVert(V7);
];	\AddBlock



proc	MakeScene;	\Set up arrays of coordinates for the scene (grinder)
\
\			    ___________
\			   /    4     /|
\			  /__________/ |
\			 /11./| /12./| |
\			/___/:|/___/:| |
\			|   |:||   |:| |
\			| 0 |:/| 1 |:/4|
\			|___|/ |___|/| |
\			  |___	 ____| |
\			 /13./| /14./| |
\			/___/:|/___/:| |
\			|   |:||   |:|/
\			| 2 |:/| 3 |:/
\			|___|/ |___|/
\
[Faces:= 0;  Verts:= 0;
\	  Coordinates of opposite corners of blocks   Colors of opposite faces
AddBlock([-60.0, -60.0, -20.0], [-10.0, -10.0,-10.0], [26, 27, 28]); \         0
AddBlock([+60.0, -60.0, -20.0], [+10.0, -10.0,-10.0], [26, 27, 28]); \         1
AddBlock([-60.0, +60.0, -20.0], [-10.0, +10.0,-10.0], [26, 27, 28]); \         2
AddBlock([+60.0, +60.0, -20.0], [+10.0, +10.0,-10.0], [26, 27, 28]); \         3
AddBlock([-60.0, -60.0,   0.0], [+60.0, +60.0, 20.0], [26, 27, 25]); \base     4
AddBlock([-10.0, -20.0, -20.0], [+10.0, +20.0,-10.0], [66, 67, 68]); \sliders  5
AddBlock([-20.0, -10.0, -20.0], [+20.0, +10.0,-10.0], [66, 67, 68]); \         6
AddBlock([-05.0, -05.0, -20.0], [105.0, +05.0,-30.0], [55, 54, 53]); \arm      7
AddBlock([-01.0, -01.0, -30.0], [+01.0, +01.0,-31.0], [26, 27, 28]); \pivots   8
AddBlock([-01.0, -01.0, -30.0], [+01.0, +01.0,-31.0], [26, 27, 28]); \         9
AddBlock([-03.0, -03.0, -30.0], [+03.0, +03.0,-50.0], [40, 12, 42]); \knob    10
AddBlock([-60.0, -60.0, -10.0], [-15.0, -15.0,  0.0], [23, 24, 25]); \        11
AddBlock([+60.0, -60.0, -10.0], [+15.0, -15.0,  0.0], [23, 24, 25]); \        12
AddBlock([-60.0, +60.0, -10.0], [-15.0, +15.0,  0.0], [23, 24, 25]); \        13
AddBlock([+60.0, +60.0, -10.0], [+15.0, +15.0,  0.0], [23, 24, 25]); \        14
AddBlock([-15.0, -20.0, -10.0], [+15.0, +20.0,  0.0], [58, 59, 60]); \sliders 15
AddBlock([-20.0, -15.0, -10.0], [+20.0, +15.0,  0.0], [58, 59, 60]); \        16
];	\MakeScene

\===============================================================================

proc	Reset;			\Initialize variables for starting configuration
[Zoom:= 3.0;			\magnification factor
XRay:= -11_000_000;		\adjusted to be just in front of scene
Offset(0):= 0;   Offset(1):= 0;

Ref0(0):= 1.0;   Ref0(1):= 0.0;   Ref0(2):= 0.0;  \tilt reference frame 30 deg
Ref1(0):= 0.0;   Ref1(1):= sqrt(3.0)/2.0;   Ref1(2):= -0.5;
];	\Reset



proc	NumOut(TLine, Num);	\Display 3-digit right-justified positive number
int	TLine, Num;		\ on the specified text line
[Cursor(2, TLine);
if Num < 100 then ChOut(6, ^ );
if Num < 10  then ChOut(6, ^ );
IntOut(6, Num);
];	\NumOut

\-------------------------------------------------------------------------------

int	FPS,			\frames per second, the display's update rate
	Sec,			\current second of time (from real-time clock)
	SecOld;			\previous second of time

[	\MoveScene
Reset;
FPS:= 0;

Cursor(0, 2);			\(device 6 always works for VESA graphic modes)
Text(6, "    Mouse rotates scene

    Left button zooms

    Right button pans

    Both buttons x-ray


      Frames/second




    Reset    [Enter]

    Exit     [Esc]");

loop	[Sec:= GetTime/1000000;	\time in seconds
	if Sec = SecOld then	\if same as before then
		FPS:= FPS+1	\ bump FPS counter
	else	[SecOld:= Sec;	\otherwise save old seconds and
		NumOut(11, FPS); \display FPS counter (once per second)
		FPS:= 0;	\reset FPS counter
		];
	MakeScene;		\with sliders and arm in new positions
	MoveSliders;
	RotateScene;
	EraseImage;
	DrawScene;
	Paint(ScrW-ImW, 0, ImW, ImH, Image, ImW);     \copy Image onto screen

	if KeyHit then
		case ChIn(1) of	\device 1 doesn't echo character and handles Esc
		  CR:	Reset;
		  Esc:	quit
		other	ChOut(0, Bel);
	];
];	\MoveScene

\###############################################################################

[	\Main
SetVid(VidMode);		\(in WinXP this affects mouse motion counters)
if not OpenMouse then Fatal("A mouse is required");
MoveScene;
SetVid(3);			\restore normal text mode for DOS prompt C:\
];	\Main
